#!/usr/bin/env python
import numpy
from sys import exit

def experiment(variables, yvar, pepa_model):
    if len(variables) == 1:
        var = variables[0]
        if var[2] == "range":
            start, stop, step = var[3].split(",")
            ran = range_maker(float(start), float(stop), float(step))
        elif var[2] == "list":
            ran = var[3].split(",")
        result = rate_experiment(var[1],ran, yvar, pepa_model)
        return result
    elif len(variables) == 2:
        var1, var2 = variables[0], variables[1]
        if var1[2] == "range":
            start, stop, step = var1[3].split(",")
            ran1 = range_maker(float(start), float(stop), float(step))
            start1, stop1, step1 = var2[3].split(",")
            ran2 = range_maker(float(start1), float(stop1), float(step1))
            result = rate_experiment_two(var1[1], ran1, var2[1], ran2, yvar, pepa_model)
            return result



def get_rate_from_actset(action, actset):
    """ Returns rate from actset returned from solver """
    for act in actset:
        if act[0] == action:
            return float(act[1])

def rate_experiment(rate_x, var_rate, rate_y, pepa_model):
    """ var_rate is a generator
        rate_y is the resulting rate on the Y axis
    """
    rate_ys = []
    rate_xs = []
    rates = pepa_model.get_rates()
    if rate_x not in rates or rate_y not in pepa_model.actions:
        print("No such rate or action")
        exit(1)
    if hasattr(var_rate, '__call__'):
        for i in var_rate():
                rates[rate_x] = str(i)
                rate_xs.append(float(i))
                pepa_model.recalculate(rates)
                pepa_model.steady_state()
                rate_ys.append( get_rate_from_actset(rate_y, pepa_model.get_throughoutput()))
    else:
        for i in var_rate:
                rates[rate_x] = str(i)
                rate_xs.append(float(i))
                pepa_model.recalculate(rates)
                pepa_model.steady_state()
                rate_ys.append( get_rate_from_actset(rate_y, pepa_model.get_throughoutput()))

    return (rate_xs, rate_ys)

def rate_experiment_two(rate_x, var_rate, rate_y, var_rate2, rate_z, pepa_model):
    """ var_rate is a generator
        rate_y is the resulting rate on the Y axis
    """
    from itertools import product
    rate_ys = []
    rate_xs = []
    rate_zs = []
    rates = pepa_model.get_rates()
    if rate_x not in rates:
        print("No such rate {}".format(rate_x))
        exit(1)
    for i,j in product(var_rate, var_rate2):
            rates[rate_x] = str(i)
            rates[rate_y] = str(j)
            rate_xs.append(float(i))
            rate_ys.append(float(j))
            pepa_model.recalculate(rates)
            pepa_model.steady_state()
            rate_zs.append( get_rate_from_actset(rate_z, pepa_model.get_throughoutput()))
    return (rate_xs, rate_ys, rate_zs)

def range_maker(low, hi, step, lst=None):
    """
    Returns a generator function
    """
    return numpy.arange(low, hi, step)




#!/usr/bin/env python
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
from matplotlib.ticker import LinearLocator, FormatStrFormatter
import matplotlib.pyplot as plt
import numpy as np
from pylab import plot, ylabel, xlabel, show, savefig

def plot_2d(xs, ys, lw=1, xlab="X", ylab="Y", action="show", name="NONAME-2d"):
    plot(xs, ys, linewidth=lw)
    xlabel("{}".format(xlab))
    ylabel("{} throughoutput".format(ylab))
    if action == "show":
        show()
    else:
        savefig("{}.png".format(name))

def plot_3d(xs, ys, zs, lw=1, xlab="X", ylab="Y", zlab="Z", action="show", name="NONAME-3d"):
    fig = plt.figure()
    ax = fig.gca(projection='3d')
    X, Y = np.meshgrid(xs, ys)
    surf = ax.plot_surface(X, Y, zs, rstride=1, cstride=1)
    plt.show()







#!/usr/bin/env python
__author__= "Dariusz Dwornikowski"
__email__ = "dariusz.dwornikowski@cs.put.poznan.pl"

""" Main file for pyPEPA """

from pprint import pprint
from pypepa import PEPAModel
from pypepa.utils import pretty_print_vector, pretty_print_performance
from pypepa.experiments.experiment import rate_experiment, range_maker,\
                                            rate_experiment_two
from pypepa.experiments.graphing import plot_2d, plot_3d
from pypepa.logger import init_log
from pypepa import __version__
import argparse
import sys


def main():
    parser = argparse.ArgumentParser(description="pypepa v{},"
                                     " author {}, {}".format(__version__,
                                     __author__, __email__))
    gen_args = parser.add_argument_group("General", "General arguments")
    gen_args.add_argument("--log", action="store", dest="loglevel", 
                          choices=["DEBUG", "INFO", "ERROR", "NONE"], 
                          help="logging level",
                          default="ERROR", type=str)
    sol_args = parser.add_argument_group("Solution",
                                         "Solution related commands")
    exp_args = parser.add_argument_group("Experimentations",
                                         "Experimentations")

    sol_args.add_argument("-s", "--solver", action="store",
                          dest="solver",type=str,
                          choices=['direct', 'sparse'],
                          help="choose solver type DEFAULT: sparse",
                          default="sparse")
    output_args = parser.add_argument_group("Output", "Output based options")
    parser.add_argument("file", help="path to the model file")
    output_args.add_argument("-gd", "--generate_dots",
                             help="generate a Graphviz dot file for every"
                                  "sequential component in a GENDOTS folder.",
                             action="store", dest="gendots", type=str)
    output_args.add_argument("-st", "--steady",
                             help="print steady state probability vector",
                             action="store_true")
    output_args.add_argument("-th", "--performance",
                             help="print throughoutput of actions",
                             action="store_true", dest="top")
    output_args.add_argument("-tr", "--transient",
                             help="print throughoutput of actions",
                             action="store", dest="trantime", type=int)
    # output_args.add_argument("-ut", "--utilization",
                               # help="print utilization of action",
                               # action="store_true", dest="util")
    output_args.add_argument("-f", "--format", dest="format", type=str,
                             choices=["graph", "console", "csv"],
                             help="format for -st -th -varrate", default="console")
    output_args.add_argument("-o", "--output", dest="output", type=str,
                               action="store",
                              help="output file valid when format cvs")
    exp_args.add_argument("-var", "--variable",
                          help="more or one variables in format"
                               "rate:RATENAME:r:START,STOP,STEP"
                               "or rate:RATENAME:l:val1,val2,val3",
                          action="append", dest="variables")
    exp_args.add_argument("-val", "--value", action="store", dest="yvar") 
    exp_args.add_argument("--actionth",
                          help="throughoutput of action on the Y axis",
                          dest="actionth", action="store",
                          metavar="action name")

    args = parser.parse_args()

    logger = init_log(log_level=args.loglevel)
    pargs = {"file": args.file, "solver" : args.solver}
    if args.gendots:
        pm = PEPAModel(**pargs)
        import os
        if os.path.isdir(args.gendots):
            pass
        else:
            os.makedirs(args.gendots)
        pm.generate_dots(args.gendots)
        sys.exit(0)

    if args.variables and args.yvar:
        variables = decode_variables(args.variables)
        if len(variables) == 1:
            pm = PEPAModel(**pargs)
            pm.derive()
            result = experiment(variables, args.yvar, pm)
            if args.format == "graph":
                plot_2d(result[0], result[1], lw=2, action="show",
                        xlab=args.yvar, ylab=variables[0].varval)
            elif args.format == "csv":
                with open("{}.csv".format(args.output), "w") as f:
                    for i in range(0, len(result[0])):
                            f.write("{}, {}\n".format(result[0][i], result[1][i]))
        elif len(variables) == 2:
            pm = PEPAModel(**pargs)
            pm.derive()
            result = experiment(variables, args.yvar, pm)
            plot_3d(result[0], result[1], result[2], action="show",
                    xlab=args.yvar, ylab=variables[0].varval)
        else:
            print("Wrong number of -var, either one or two")
            sys.exit(1)

    pm = PEPAModel(**pargs)


    if args.steady or args.top:
        pm.steady_state()
        print("Statespace of {} has {} states \n".format(args.file,
              len(pm.get_steady_state_vector() )))
    if args.trantime:
        tr = pm.transient(0, int(args.trantime))
        print("Transient analysis from time %d to %d" % (0, args.trantime))
        args.output = "{}-transient.csv".format(pm.name)
        pretty_print_vector(tr,
                             pm.get_state_names(),
                             fmt=args.format,
                             outfile=args.output
                             )
    if args.steady:
        print("Steady state vector")
        args.output = "{}-steady.csv".format(pm.name)
        pretty_print_vector(pm.get_steady_state_vector(),
                             pm.get_state_names(),
                             fmt=args.format,
                             outfile=args.output
                             )
    if args.top:
        print("Throuhoutput (successful action completion in one time unit)")
        print("Output:{}".format(args.format))
        args.output = "{}-throughput.csv".format(pm.name)
        pretty_print_performance(pm.get_throughoutput(), fmt=args.format,
                                                         outfile=args.output)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

class ComponentStateVisitor():

    def __init__(self, graph, output_dir = "dots"):
        self.graph = graph
        self.visited = []
        self.dot = ""
        self.out_dir = output_dir

    def generate_ss(self, node, comp):
        self.visited = []
        return self._visit_for_ss(node,comp)

    def _visit_for_ss(self, node, comp):
        self.visited.append(node)
        comp.ss[node] = self.graph.ss[node]
        transitions = self.graph.ss[node].transitions
        for tran in transitions:
            if tran.action in self.graph.shared_actions:
                comp.shared.append( (tran.action, tran.rate) )
            comp.activities.append( (tran.action, tran.rate) )
            if tran.to not in self.visited:
                self._visit_for_ss(tran.to, comp)
        return comp


    def get_dot(self, node):
        with open(self.out_dir + "/"+node + ".dot", "w") as f:
            self.dot = "digraph %s {\n" % (node)
            self._visit_dot(node)
            self.dot += "}\n"
            f.write(self.dot)
        return self.dot

    def _visit_dot(self, node):
        self.visited.append(node)
        transitions = self.graph.ss[node].transitions
        for tran in transitions:
            if tran.action in self.graph.shared_actions:
                self.dot += "\"%s\" -> \"%s\" [label=\"(%s,%s)\"" \
                            "fontsize=10]\n" % (node,tran.to,
                                                tran.action, str(tran.rate))
            else:
                self.dot += "\"%s\" -> \"%s\" [label=\"(%s,%s)\"" \
                            "fontsize=10]\n" % (node, tran.to,
                                                tran.action, str(tran.rate))
            if tran.to not in self.visited:
                self._visit_dot(tran.to)





___author__ = "Dariusz Dwornikowski"
__copyright__ = "Dariusz Dwornikowski, Poznan University of Technology"
__email__ = "dariusz.dwornikowski@cs.put.poznan.pl"
__version__ = "0.4.6dev"

from pypepa.pepa_model import PEPAModel



#!/usr/bin/env python
"""
Module with classes for PEPA model.

"""

class BaseNode(object):
    left, right = None, None
    data = None
    asttype = None
    length = None

    def __init__(self, data, asttype):
        self.data = data
        self.asttype = asttype

    def __str__(self):
        return self.data

class ChoiceNode(BaseNode):
    lhs, rhs = None, None
    reolved = None

    def __init__(self, data, asttype):
        super(ChoiceNode, self).__init__(data, asttype)

class PrefixNode(BaseNode):
    action, resolved, rate, var_rate  = None, None, None, None

    def __init__(self, data, asttype):
        super(PrefixNode, self).__init__(data, asttype)

class DefNode(BaseNode):
    process, resolved = None, None

    def __init__(self, data, asttype):
        super(DefNode, self).__init__(data, asttype)

class ActivityNode(BaseNode):

    action, rate = "", ""

    def __init__(self, data, asttype):
        super(ActivityNode, self).__init__(data, asttype)

class ProcdefNode(BaseNode):
    name  = None
    aggregation = False
    aggr_num = 0

    def __init__(self, data, asttype):
        super(ProcdefNode, self).__init__(data, asttype)

class CoopNode(BaseNode):
    cooptype, actionset = None, None

    def __init__(self, data, asttype):
        super(CoopNode, self).__init__(data, asttype)

class SyncsetNode(BaseNode):
    actionset = None

    def __init__(self, data, asttype):
        super(SyncsetNode, self).__init__(data, asttype)




#!/usr/bin/env python
""" 
This parser is inspired by fourFn.py example by Paul McGuire. Thank You Paul
for amazing pyparsing library.
"""
from pyparsing import Literal,CaselessLiteral,Word,Combine,Group,Optional,\
        ZeroOrMore,Forward,nums,alphas, alphanums
import readline
import operator
import re

class RateParser(object):
    """
    Parses and evaluates expresssions mathematical
    expressions, ops supported: +, -, *, /, ^
    """
    def __init__(self):
        self.bnf = None
        self.expr_stack = []
        self.variables = {}
        self.var_stack = []
        self.opn = { "+" : operator.add,
                "-" : operator.sub,
                "*" : operator.mul,
                "/" : operator.truediv,
                "^" : operator.pow }

    def _pushFirst(self, string, loc, toks):
        self.expr_stack.append(toks[0])

    def _assignVar(self, str, loc, toks):
        self.var_stack.append(toks[0])

    def grammar(self):
        if not self.bnf:
            point = Literal( "." )
            fnumber = Combine( Word( nums ) + 
                               Optional( point + Optional( Word( nums ) ) ) 
                               )
            ident = Word(alphas.lower()+"_", alphanums+"_")
            plus  = Literal( "+" )
            minus = Literal( "-" )
            mult  = Literal( "*" )
            # passiverate = Word('infty') | Word('T')
            div   = Literal( "/" )
            lpar  = Literal( "(" ).suppress()
            rpar  = Literal( ")" ).suppress()
            addop  = plus | minus
            multop = mult | div
            assign = Literal('=')
            expop = Literal( "^" )
            expr = Forward()
            atom = Optional("-") + ( fnumber | ident + lpar + expr + rpar | ident).setParseAction(self._pushFirst ) | lpar + expr.suppress() + rpar 
            factor = Forward()
            factor << atom + ZeroOrMore( ( expop + factor )
                    .setParseAction(self._pushFirst) )
            term = factor + ZeroOrMore( ( multop + factor )
                    .setParseAction(self._pushFirst) )
            expr << term + ZeroOrMore( ( addop + term )
                    .setParseAction(self._pushFirst ) )
            bnf = (ident + assign).setParseAction(self._assignVar) + expr 
            self.bnf = bnf
        return self.bnf


    def parse_rate_expr(self, rate):
        result = self.evaluate(self.expr_stack)
        self.expr_stack = []
        if len(self.var_stack) == 1:
            self.variables[self.var_stack.pop()] = result
        return result


    def evaluate(self,s):
        op = s.pop()
        if op in self.opn:
            op2 = self.evaluate(s)
            op1 = self.evaluate(s)
            return self.opn[op](op1, op2)
        elif  re.search('^[a-z][a-zA-Z0-9_]*$',op):
            if op in self.variables:
                return self.variables[op]
            else:
                raise Exception("Variable {} not defined".format(op))
        # elif op[0].isalpha():
        #     return 0
        else:
            return float( op )

# if __name__ == "__main__":
#     parser = RateParser()
#     input_string = ''
#     input_string = input("> ")
#     while input_string != 'quit':
#         if input_string != '':
#             result = parser.parse_rate_expr(input_string)
#         input_string = input("> ")






#!/usr/bin/env python
"""
PEPA Parser
"""
from pyparsing import Word, Literal, alphas, alphanums, nums, Combine, Optional, ZeroOrMore, Forward, restOfLine
import sys
from parsing.pepa_ast import *

_logging_pa = False
_var_stack = None
_processes = None
_systemeq = None

def __init__(logging_pa = False):
    _logging_pa = logging_pa
    _processes = {}
    _var_stack = {}

def log_pa(string, msg="", prepend="[PARSEACT]"):
    fname = sys._getframe(1).f_code.co_name
    if _logging_pa:  print(prepend + "[" + fname + "]", msg, string)

def error(string):
    print("SYNTAX ERROR: ", string)


def _create_activity(string, loc,tok):
    log_pa("Token: "+tok[0])
    n = ActivityNode("(" + tok[0] + "," + tok[1] + ")", "activity")
    n.action = tok[0]
    n.rate = tok[1]
    return n

def _create_procdef(string, loc, tok):
    log_pa("Token: "+tok[0])
    n = ProcdefNode(tok[0], "procdef")
    if len(tok) > 1:
        n.aggregation = True
        if tok[1] in _var_stack:
            n.aggr_num = int(_var_stack[tok[1]])
        else:
            n.aggr_num = int(tok[1])
    n.name = tok[0]
    return n

def _create_definition(string, loc, tok):
    log_pa("Left token: "+tok[0].data)
    log_pa("Right token: "+tok[2].data)
    n = DefNode("=", "definition")
    n.left = tok[0]
    n.right = tok[2]
    n.process = tok[0].data
    for key in _processes.keys():
        if _processes[key].process == tok[0].data:
            error("Process "+tok[0].data+" already defined")
            exit(1)
    _processes[tok[0]] = n
    return n

def _create_prefix(string, loc, tok):
    log_pa("Tokens: "+str( len(tok) ))
    if len(tok) > 1:
        log_pa("Left token: "+tok[0].data)
        log_pa("Right token: "+tok[2].data)
        n = PrefixNode(".", "prefix")
        lhs = tok[0]
        rhs = tok[2]
        n.left = lhs
        n.right = rhs
        return n
    else:
        log_pa("Token: "+tok[0].data)
        return tok[0]

def _create_choice(string,loc,tok):
    log_pa("Start")
    log_pa("Tokens: "+str( len(tok) ))
    if not tok[0] is None:
        if len(tok) <3:
            log_pa("Token: "+tok[0].data)
            return tok[0]
        else:
            log_pa("Left token: "+tok[0].data)
            log_pa("Right token: "+tok[2].data)

            n = ChoiceNode("+", "choice")
            n.left = tok[0]
            n.right = tok[2]
            return n


def _create_coop(string, loc, tok):
    if not tok[0] is None:
        if len(tok) <3:
            log_pa("Token: "+tok[0].data)
            return tok[0]
        else:
            log_pa("Number of tokens" + str(len(tok)))
            log_pa("Left token: "+tok[0].data)
            if tok[1].actionset is not None:
                n = CoopNode("<"+str(tok[1].actionset)+">", "coop")
                n.cooptype = "sync"
                n.actionset = tok[1].actionset
                log_pa(n.actionset)
            else:
                n = CoopNode("||", "coop")
                n.cooptype = "par"
            if type(tok[2]).__name__ == "str":
                log_pa("String: "+tok[2])
            else:
                log_pa("Right token: "+tok[2].data)
            n.left = tok[0]
            n.right= tok[2]
            return n

def _create_sync_set(string, loc, tok):
    if tok[0] != "||" and tok[0] != "<>":
        log_pa("Non empty synset: " + str(tok))
        n = SyncsetNode("<>", "syncset")
        n.actionset = tok
    else:
        log_pa("Parallel")
        n = SyncsetNode("||", "syncset")
        n.actionset = None
    return n

def _create_subtree_aggregation(num, procname):
    """ Transforms Process[num] into AST subtree """
    first = CoopNode("||", "coop")
    first.cooptype = "par"
    last = first
    for i in list(range(2, num+1)):
        nl = ProcdefNode(procname, "procdef")
        if i == num:
            nr = ProcdefNode(procname, "procdef")
            last.right = nr
        else:
            nr = CoopNode("||", "coop")
            nr.cooptype = "par"
            last.right = nr
        last.left = nl
        last = nr
    return first

def _create_process(string, loc, tok):
    log_pa("Start")
    if tok[0].left is not None or tok[0].right is not None:
        log_pa("Token: "+tok[0].data)
        log_pa("Non terminal - passing")
        return tok[0]
    else:
        if tok[0].asttype == "procdef":
            n = ProcdefNode(tok[0].data, tok[0].asttype)
            # TODO create subtree
            if tok[0].aggregation == True:
                n.aggregation = True
                n.aggr_num = tok[0].aggr_num
                n = _create_subtree_aggregation(tok[0].aggr_num, tok[0].data)
                n.asttype = "coop"
                n.data = "||"
                n.cooptype = "par"
                n.actionset = None
        elif tok[0].asttype == "activity":
            n = ActivityNode(tok[0].data, tok[0].asttype)
            n.rate = tok[0].rate
            n.action = tok[0].action
        else:
            log_pa("ERRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRROR")
        log_pa("Terminal - creating Node ->" + tok[0].asttype)
        log_pa("Token: "+tok[0].data)
    return n


def _create_system_equation(string, loc, tok):
    log_pa("Creating system EQ")
    _systemeq = tok[0]


def assign_var(toks):
    _var_stack[toks[0]] = toks[2]

def _check_var(str,loc,tok):
    try:
        # just a number
        float(tok[0])
    except:
        try:
            if tok[0] not in ("infty", "T", "tau"):
                _var_stack[tok[0]]
        except:
            error("Rate " + tok[0]+ " not defined")
            exit(1)

def gramma():
## Tokens
    point = Literal('.')
    prefix_op = Literal('.')
    choice_op = Literal('+')
    parallel = Literal("||") | Literal("<>")
#ident = Word(alphas, alphanums+'_')
    ratename = Word(alphas.lower(),alphanums+"_")
    lpar = Literal('(').suppress()
    rpar = Literal(')').suppress()
    lsqpar = Literal('[').suppress()
    rsqpar = Literal(']').suppress()

    define = Literal('=')
    semicol = Literal(';').suppress()
    col = Literal(',').suppress()
    number = Word(nums)
    integer = number
    floatnumber = Combine( integer + Optional( point + Optional(number)))
    passiverate = Word('infty') | Word('T')
    internalrate = Word('tau')
    pound = Literal('#').suppress()
    percent = Literal('%').suppress()
    peparate = (ratename | floatnumber | internalrate | passiverate).setParseAction(_check_var)
    peparate_indef = floatnumber | internalrate | passiverate
    sync = Word('<').suppress() + ratename + ZeroOrMore(col + ratename) + Word('>').suppress()
    coop_op = (parallel | sync).setParseAction(_create_sync_set)
    activity = (ratename + col + peparate).setParseAction(_create_activity)
    procdef = (Word(alphas.upper(), alphanums+"_") + Optional(lsqpar + peparate_indef + rsqpar)).setParseAction(_create_procdef)
## RATES Definitions
    ratedef = (Optional(percent)+ratename + define + peparate_indef).setParseAction(assign_var) + semicol

    prefix = Forward()
    choice = Forward()
    coop = Forward()

    process = ( activity
             | procdef
             | lpar + coop + rpar
            ).setParseAction(_create_process)
    prefix  << (process + ZeroOrMore(prefix_op + prefix)).setParseAction( _create_prefix)
    choice << (prefix + ZeroOrMore(choice_op + choice)).setParseAction(_create_choice)
    coop << (choice + ZeroOrMore(coop_op + coop)).setParseAction(_create_coop)
    rmdef = (Optional(pound) + procdef + define + coop + semicol).setParseAction(_create_definition)
    system_eq =  Optional(pound) + coop
    pepa = ZeroOrMore(ratedef)  + ZeroOrMore(rmdef) + system_eq.setParseAction(_create_system_equation)
    pepacomment = '//' + restOfLine
    pepa.ignore(pepacomment)
    return pepa

def parse(string):
    gramma().parseString(string)
    return (_processes, _var_stack, _systemeq)






#!/usr/bin/env python
from pyparsing import Word, Literal, alphas, alphanums, Combine, Optional,\
                      ZeroOrMore, Forward, restOfLine, nums, lineno, col, ParseException
from pypepa.parsing.pepa_ast import *
from pypepa.parsing.rate_parser import RateParser
from pypepa.logger import init_log

class PEPAParser(object):

    def __init__(self):
        self.log_pa = init_log()
        self._processes = {}
        self._var_stack = {}
        self._actions = []
        self._seen = {}
        self.systemeq = None
        self.rate_parser = RateParser()

    def _create_activity(self, string, loc,tok):
        self.log_pa.debug("Token: "+tok[0])
        n = ActivityNode("(" + tok[0] + "," + tok[1] + ")", "activity")
        n.action = tok[0]
        self._actions.append(n.action)
        n.rate = tok[1]
        return n

    def _create_procdef(self, s, loc, toks):
        self.log_pa.debug("Token: "+toks[0])
        n = ProcdefNode(toks[0], "procdef")
        self._seen[toks[0]] = (loc, s)
        if len(toks) > 1:
            n.aggregation = True
            if toks[1] in self._var_stack:
                n.aggr_num = int(self._var_stack[toks[1]])
            else:
                n.aggr_num = int(toks[1])
        n.name = toks[0]
        return n

    def _create_definition(self, string, loc, tok):
        self.log_pa.debug("Left token: "+tok[0].data)
        self.log_pa.debug("Right token: "+tok[2].data)
        n = DefNode("=", "definition")
        n.left = tok[0]
        n.right = tok[2]
        n.process = tok[0].data
        for key in self._processes.keys():
            if self._processes[key].process == tok[0].data:
                self.log_pa.error("Process "+tok[0].data+" already defined")
                raise Exception("Process {} already defined".format(tok[0].data))
        self._processes[tok[0]] = n
        return n

    def _create_prefix(self,string, loc, tok):
        self.log_pa.debug("Tokens: "+str( len(tok) ))
        if len(tok) > 1:
            self.log_pa.debug("Left token: "+tok[0].data)
            self.log_pa.debug("Right token: "+tok[2].data)
            n = PrefixNode(".", "prefix")
            lhs = tok[0]
            rhs = tok[2]
            n.left = lhs
            n.right = rhs
            return n
        else:
            self.log_pa.debug("Token: "+tok[0].data)
            return tok[0]

    def _create_choice(self,string,loc,tok):
        self.log_pa.debug("Tokens: "+str( len(tok) ))
        if not tok[0] is None:
            if len(tok) <3:
                self.log_pa.debug("Token: "+tok[0].data)
                return tok[0]
            else:
                self.log_pa.debug("Left token: "+tok[0].data)
                self.log_pa.debug("Right token: "+tok[2].data)

                n = ChoiceNode("+", "choice")
                n.left = tok[0]
                n.right = tok[2]
                return n

    def _create_coop(self,string, loc, tok):
        if not tok[0] is None:
            if len(tok) <3:
                self.log_pa.debug("Token: "+tok[0].data)
                return tok[0]
            else:
                self.log_pa.debug("Number of tokens" + str(len(tok)))
                self.log_pa.debug("Left token: "+tok[0].data)
                if tok[1].actionset is not None:
                    n = CoopNode("<"+str(tok[1].actionset)+">", "coop")
                    n.cooptype = "sync"
                    n.actionset = tok[1].actionset
                    self.log_pa.debug(n.actionset)
                else:
                    n = CoopNode("||", "coop")
                    n.cooptype = "par"
                if type(tok[2]).__name__ == "str":
                    self.log_pa.debug("String: "+tok[2])
                else:
                    self.log_pa.debug("Right token: "+tok[2].data)
                n.left = tok[0]
                n.right= tok[2]
                return n

    def _create_sync_set(self,string, loc, tok):
        if tok[0] != "||" and tok[0] != "<>":
            self.log_pa.debug("Non empty synset: " + str(tok))
            n = SyncsetNode("<>", "syncset")
            n.actionset = tok
        else:
            self.log_pa.debug("Parallel")
            n = SyncsetNode("||", "syncset")
            n.actionset = None
        return n

    def _create_subtree_aggregation(self, num, procname):
        """ Transforms Process[num] into AST subtree """
        first = CoopNode("||", "coop")
        first.cooptype = "par"
        last = first
        for i in range(2, num+1):
            nl = ProcdefNode(procname, "procdef")
            if i == num:
                nr = ProcdefNode(procname, "procdef")
                last.right = nr
            else:
                nr = CoopNode("||", "coop")
                nr.cooptype = "par"
                last.right = nr
            last.left = nl
            last = nr
        return first

    def _create_process(self,string, loc, tok):
        self.log_pa.debug("Start")
        if tok[0].left is not None or tok[0].right is not None:
            self.log_pa.debug("Token: "+tok[0].data)
            self.log_pa.debug("Non terminal - passing")
            return tok[0]
        else:
            if tok[0].asttype == "procdef":
                n = ProcdefNode(tok[0].data, tok[0].asttype)
                if tok[0].aggregation == True and n.aggr_num > 1:
                    n.aggregation = True
                    n.aggr_num = tok[0].aggr_num
                    n = self._create_subtree_aggregation(tok[0].aggr_num, tok[0].data)
                    n.asttype = "coop"
                    n.data = "||"
                    n.cooptype = "par"
                    n.actionset = None
            elif tok[0].asttype == "activity":
                n = ActivityNode(tok[0].data, tok[0].asttype)
                n.rate = tok[0].rate
                n.action = tok[0].action
            else:
                self.log_pa.error("This situation should not take place")
            self.log_pa.debug("Terminal - creating Node ->" + tok[0].asttype)
            self.log_pa.debug("Token: "+tok[0].data)
        return n

    def _create_system_equation(self, string, loc, tok):
        self.log_pa.debug("Creating system EQ")
        self._systemeq = tok[0]

    def _assign_var(self,toks):
        result = self.rate_parser.parse_rate_expr("".join(toks))
        if toks[0] not in self._var_stack:
            self._var_stack[toks[0]] = str(result)
        else:
            raise Exception("Variable {} has been already defined".format(toks[0]))

    def _check_var(self,str,loc,tok):
        try:
            float(tok[0])
        except:
            pass
        else:
            return
        if tok[0] not in ("infty", "T"):
            if tok[0] not in self._var_stack:
                raise Exception("Rate {} not defined in {}".format(tok[0], loc))

    def gramma(self):
## Tokens
        point = Literal('.')
        prefix_op = Literal('.')
        choice_op = Literal('+')
        parallel = Literal("||") | Literal("<>")
        ratename = Word(alphas.lower(),alphanums+"_")
        lpar = Literal('(').suppress()
        rpar = Literal(')').suppress()
        lsqpar = Literal('[').suppress()
        rsqpar = Literal(']').suppress()

        define = Literal('=')
        semicol = Literal(';').suppress()
        col = Literal(',').suppress()
        number = Word(nums)
        integer = number
        floatnumber = Combine( integer + Optional( point + Optional(number)))
        passiverate = Word('infty') | Word('T')
        pound = Literal('#').suppress()
        percent = Literal('%').suppress()
        peparate = (ratename | floatnumber | passiverate).setParseAction(self._check_var)
        # peparate_indef = floatnumber | internalrate | passiverate
        sync = Word('<').suppress() + ratename + ZeroOrMore(col + ratename) + Word('>').suppress()
        coop_op = (parallel | sync).setParseAction(self._create_sync_set)
        activity = (ratename + col + peparate).setParseAction(self._create_activity)
        procdef = (Word(alphas.upper(), alphanums+"_"+"`"+"'") + Optional(lsqpar + peparate + rsqpar)).setParseAction(self._create_procdef)
## RATES Definitions
        # ratedef = (Optional(percent)+ratename + define + peparate_indef).setParseAction(self._assign_var) + semicol
        ratedef = Optional(percent)+(self.rate_parser.grammar()).setParseAction(self._assign_var) + semicol

        prefix = Forward()
        choice = Forward()
        coop = Forward()

        process = ( activity
                 | procdef
                 | lpar + coop + rpar
                ).setParseAction(self._create_process)
        prefix  << (process + ZeroOrMore(prefix_op + prefix)).setParseAction(self. _create_prefix)
        choice << (prefix + ZeroOrMore(choice_op + choice)).setParseAction(self. _create_choice)
        coop << (choice + ZeroOrMore(coop_op + coop)).setParseAction(self._create_coop)
        rmdef = (Optional(pound) + procdef + define + coop + semicol).setParseAction(self._create_definition)
        system_eq =  Optional(pound) + coop
        pepa = ZeroOrMore(ratedef)  + ZeroOrMore(rmdef) + system_eq.setParseAction(self._create_system_equation)
        pepacomment = '//' + restOfLine
        pepa.ignore(pepacomment)
        return pepa

    def parse(self,string):
        try:
            self.gramma().parseString(string, parseAll=True)
        except ParseException as e:
            raise
        seen_procs = [str(x) for x in self._processes]
        for seen in self._seen:
            if seen not in seen_procs:
                line = lineno(self._seen[seen][0], self._seen[seen][1])
                column =col(self._seen[seen][0], self._seen[seen][1])
                raise Exception("{} process not defined - possible deadlock, line {}, col {}".format(seen,line, column))
        return (self._processes, self._var_stack, self._systemeq, self._actions)






#!/usr/bin/env python

from pypepa.logger import init_log

class StateSpace():
    operators = []
    components = []
    comp_ss = None


    def __init__(self):
        self.max_length = 0
        self.log = init_log()

    def _combine_states(self, states):
        """ When more than one transition leads to the same state,
            the function combines these transitions into one with actions rewritten
        """
        for i in list( range(0, len(states))):
            if states[i] is not None and len(states[i].to_s) == 1:
                continue
            if states[i] is not None:
                for j in list (range(i+1, len(states))):
                    if states[j] is not None:
                        if states[i].to_s == states[j].to_s:
                            states[i].actions = {}
                            states[i].actions[states[i].action] = states[i].rate
                            states[i].combined = True
                            states[i].action += "," + states[j].action
                            states[i].rate =  float(states[i].rate) + float(states[j].rate)
                            states[i].actions[states[j].action] = float(states[j].rate)
                            states[j] = None
        states = [i for i in states if i is not None]
        return states


    def derive(self):
        """ Derives the whole state space according to BU"""
        initial_state = []
        queue = []
        visited = []
        resulting_states = {}
        actions_to_state = {}
        state_num = 0
        for x in list(range(0,self.max_length+1,1)):
            for op in self.operators:
                if op.length == x:
                    op.update_offset()
        for comp in self.components:
            initial_state.append(comp.name)
        queue.append(initial_state)
        while(queue):
            state = queue.pop(0)
            if self._gs_to_string(state) in visited:
              continue
            state_num = state_num + 1
            self.log.debug("{} STATE {}".format(state_num, state))
            resulting_states[self._gs_to_string(state)] = ([],state_num)
            visited.append(self._gs_to_string(state))
            # update components table (same refs are in operators) -->
            # new state
            for i in list( range(0, len( state ),1)):
                self.components[i].update( state[i] )
            # for every length can be changed
            for x in list(range(0,self.max_length+1,1)):
                for op in self.operators:
                    if op.length == x:
                        # Runs through all operators, if operator is the
                        # top one it composes
                        if self.max_length == op.length:
                            new_states = []
                            new_states = op.compose(self.comp_ss, state, True)
                            if not new_states:
                                self.log.error("DEADLOCK in {}".format(state))
                                raise Exception("Deadlock in state: {}".format(state))
                            new_states = self._combine_states(new_states[:])
                            for news in new_states:
                                # aggregate
                                if len(news.to_s) < self.max_length:
                                    new_state = state[:]
                                    new_state[news.offset] = news.to_s[0]
                                    news.to_s = new_state
                                self.log.debug("{}\trate={} \t-> {}".format(news.action, news.arate, news.to_s))
                                resulting_states[self._gs_to_string(state)][0].append( (news.rate, self._gs_to_string(news.to_s)))
                                #handle combines actions, not very elegant so to be changed
                                if news.combined:
                                    for act in news.actions:
                                        self._add_to_actions_set(act, news.actions[act], actions_to_state, state_num)
                                else:
                                    self._add_to_actions_set(news.action, news.rate, actions_to_state, state_num)

                                if self._gs_to_string(news.to_s) not in visited:
                                    queue.append(news.to_s)
                        else:
                            op.compose(self.comp_ss , state , False)
        return (resulting_states, actions_to_state)

    def _add_to_actions_set(self, action, rate,actions_to_state, state_num):
        if rate == "infty":
            self.log.error("Resulting rate cannot be infty in {} in state {}".format(action, state_num))
            raise Exception("Resulting rate cannot be infty in {} in state {}".format(action, state_num))
        if (action,state_num) not in actions_to_state:
            actions_to_state[ (action, state_num) ] = float(rate)
        else:
            #check if the same stateand action, so we add rates
            actions_to_state[ (action, state_num) ] += float(rate)

    def _gs_to_string(self, gs_list):
        return ','.join( map( str, gs_list ) )

class Component():
    length = None
    offset = None
    name = None
    ss = None
    data = None


    def update(self, name):
        """
        Update with new name
        """
        self.name
        self.name = name
        self.data = name #TODO: wyjebac
        self.derivatives = []
        arates = {}
        for der in self.ss[self.name].transitions:
            if der.action not in arates:
                if der.rate == "infty":
                    arates[der.action] = float(-1)
                else:
                    arates[der.action] = float(der.rate)
            else:
                arates[der.action] += float(der.rate)
        for der in self.ss[self.name].transitions:
            # print("self.name:%s der.action:%s der.rate:%s de.to:%s" % (self.name, der.action, der.rate, der.to))
            if der.rate == "infty":
                der.rate = -1
            self.derivatives.append(Derivative(self.name, [der.to], der.action, der.rate, self.offset, arates[der.action], aprates=arates))



    def __init__(self, ss, name,offset):
        self.name = name
        self.ss = ss
        self.offset = offset
        self.derivatives = []
        arates = {}
        for der in self.ss[self.name].transitions:
            if der.action not in arates:
                if der.rate == "infty":
                    arates[der.action] = float(-1)
                else:
                    arates[der.action] = float(der.rate)
            else:
                arates[der.action] += float(der.rate)

        for der in self.ss[self.name].transitions:
            if der.rate == "infty":
                der.rate = -1
            self.derivatives.append(Derivative(self.name, [der.to], der.action, der.rate, self.offset, der.rate, aprates=arates))

    def get_derivatives(self):
        return self.derivatives

    def __str__(self):
        return self.name


class Derivative():

    def __init__(self, from_s, to_s, action, rate, offset, arate, shared=False, aprates=None):
        self.from_s = from_s
        self.to_s = to_s
        self.action = action
        self.rate = rate
        self.shared = shared
        self.offset = offset
        self.combined = False
        self.arate = arate
        self.apparent_rates = aprates

    def __str__(self):
        return " T:" + str(self.to_s) \
                + " Act:"+self.action+" R:"+self.rate+" O:"+str(self.offset)+" Sh:"+str(self.shared)

class Operator(Component):
    length = None
    offset = 0
    actionset = []
    lhs = None
    rhs = None

    def __init__(self):
        self.actionset = []
        self.derivatives = []
        self.log = init_log()

    def update_offset(self):
        if self.lhs is not None:
            self.offset = self.lhs.offset

    def get_derivatives(self):
        return self.derivatives

    def __str__(self):
        return "<" + str( self.actionset ) + ">"

    def _create_shared_trans(self, state, tran_l, tran_r):
        to_state = state[:]
        for i in list(range(0, self.lhs.length)):
            to_state[self.lhs.offset + i] = tran_l.to_s[self.lhs.offset + i]
        for i in list(range(0, self.rhs.length)):
            to_state[self.rhs.offset + i] = tran_r.to_s[self.rhs.offset + i]
#            to_state[i] = tran_r.to_s[i]
        left = float(tran_l.rate) / float(tran_l.arate)
        right = float(tran_r.rate) / float(tran_r.arate)
        lR = float(left) * float(right)
        new_rate = lR * float(self._min(tran_r.arate, tran_l.arate))
        # print("%s %s %s %s"% (left, right, lR, new_rate))
        ddd = Derivative(state, to_state,tran_l.action,new_rate,self.offset, new_rate,True)
        return ddd

    def _create_left_transition(self, state, tran_l):
        new_state = state[:]
        new_state[tran_l.offset] = tran_l.to_s[0]
        return tran_l

    def _create_right_transition(self, state, tran_r):
        new_state = state[:]
        new_state[tran_r.offset] = tran_r.to_s[0]
        return tran_r

    def compose(self,ss, state, topop=False):
        self.derivatives =  []
        for tran_l in self.lhs.get_derivatives():
            if tran_l.action not in self.actionset:
                for tran_r in self.rhs.get_derivatives():
                    if tran_r.action not in self.actionset:
                        if tran_l.action == tran_r.action:
                            # print("%s %s %s %s" % (tran_l.action, tran_r.action, tran_l.arate, tran_r.arate))
                            new_arate = tran_l.arate + tran_r.arate
                            tran_l.apparent_rates[tran_l.action] = new_arate
                            tran_r.apparent_rates[tran_l.action] = new_arate
                            tran_l.arate = new_arate
                            tran_r.arate = new_arate
        for tran_l in self.lhs.get_derivatives():
            # UNSHARED
            if tran_l.action not in self.actionset:
                l_tran = self._create_left_transition(state, tran_l)
                self.derivatives.append(l_tran)
            else:
                for tran_r in self.rhs.get_derivatives():
                    if tran_r.action == tran_l.action:
                        new_state = state[:]
                        if len(tran_r.to_s) == 1:
                            new_tran_s = new_state[:]
                            new_tran_s[tran_r.offset] = tran_r.to_s[0]
                            tran_r.to_s = new_tran_s
                        if len(tran_l.to_s) == 1:
                            new_tran_s = new_state[:]
                            new_tran_s[tran_l.offset] = tran_l.to_s[0]
                            tran_l.to_s = new_tran_s
                        ddd = self._create_shared_trans(state, tran_l, tran_r)
                        self.derivatives.append(ddd)
        for tran_r in self.rhs.get_derivatives():
            if tran_r.action not in self.actionset:
                self.derivatives.append(tran_r)
        if topop == True:
            return self.derivatives

    def _min(self,rate1, rate2):
        if rate1 < 0:
            if rate2 < 0:
                return -1
            else:
                return rate2
        if rate2 < 0:
            if rate1 < 0:
                return -1
            else:
                return rate1
        return min(rate1,rate2)


    def _min2(self,rate1, rate2):
        if rate1 == "infty":
            if rate2 == "infty":
                return "infty"
            else:
                return rate2
        if rate2 == "infty":
            if rate1 == "infty":
                return "infty"
            else:
                return rate1
        return min(rate1,rate2)





#!/usr/bin/env python

class ComponentState():

    def __init__(self):
        self.name = None
        self.resolved = None
        self.transitions = []
        self.apparent_rates = {}

class Transition():

    def __init__(self, action, rate, to):
        self.rate = rate
        self.var_rate = None
        self.action = action
        self.to = to


class SystemEquation():

    def __init__(self):
        self.processes = {}


class ComponentSSGraph():

    def __init__(self, name=""):
        self.name = name
        self.ss = {}
        self.activities = []
        self.shared = []

    def __str__(self):
        return "Component %s %d" % (self.name, len(self.ss.keys()))


class ModelSSGraph():
    """
    self.ss is a hash table, the keys are state names
    e.g. P1 or resolved where not possible -> ComponenStates
    """
    def __init__(self):
        self.name = ""
        self.ss = {}
        self.dotrepr = ""
        self.shared_actions = None





#!/usr/bin/env python
from collections import namedtuple
import sys

_vartup = namedtuple("VarTup", "vartype, varval, modtype, modval")

# rate:nazwa:mode:costam
def decode_variables(variables):
    varlist = []
    for var in variables:
        split = var.split(":")
        if split[0] != "rate" or split[2] not in ["list", "range"]:
            print("Invalid arguments")
            sys.exit(1)
        varlist.append(_vartup(split[0], split[1], split[2], split[3]))
    return varlist

def pretty_print_performance(actset, fmt="console", outfile="throughoutput.csv"):
    if fmt == "console" or fmt == "graph":
        for perf in actset:
            print("{0:<40} {1:>10}".format(perf[0],perf[1]) )
    elif fmt == "csv":
        with open(outfile, "w") as f:
            f.write("Rate name;Rate throughput\n")
            for perf in actset:
                f.write("{};{}\n".format(perf[0],perf[1]))


def pretty_print_vector(vect, vect_names, fmt="console", outfile="steady.csv"):
    if fmt == "console" or fmt == "graph":
        for i, prob in enumerate(vect):
            print("{};{};{}".format(i+1, vect_names[i], vect[i]))
    elif fmt == "csv":
        with open(outfile, "w") as f:
            f.write("State number;State name;result\n")
            for i, prob in enumerate(vect):
                f.write("{};{};{}\n".format(i+1, vect_names[i], vect[i]))






import logging
import sys

_levels = {"INFO": logging.INFO, "DEBUG": logging.DEBUG, "ERROR" :logging.ERROR}
logger = None

def init_log(log_level="ERROR", logger_name="pypepa"):
    global logger
    if logger is not None:
        return logger
    if log_level == "NONE":
        logging.basicConfig(level=logging.INFO,
                        format="%(name)s:%(levelname)s:%(filename)s: %(funcName)s:%(lineno)d  -  %(message)s")
        log = logging.getLogger(logger_name)
        log.setLevel(logging.INFO)
        log.disabled = True
        logger = log
    else:
        logging.basicConfig(level=_levels[log_level],
                        format="%(name)s:%(levelname)s:%(filename)s: %(funcName)s:%(lineno)d  -  %(message)s")
        log = logging.getLogger(logger_name)
        log.setLevel(log_level)
        logger = log
    return log



#!/usr/bin/env python
from pypepa.solvers.ctmc import ctmc, ctmc_sparse, create_matrix, \
                                 vector_mult, create_lil_matrix, ctmc_transient

class CTMCSolution():

    def __init__(self, ss, solver):
        self._ss = ss
        self._res = None
        self._actset = None
        self._solver = solver
        self._steady_state_vector = None


    def solve_transient(self,start, stop):
        (self._res, self._actset) = self._ss.derive()
        matrix = create_matrix(self._res)
        a = ctmc_transient(matrix, len(self._res),start, stop)
        self._vect_names = self._prepare_new_vector(self._res)
        return a

    def solve_steady(self):
        (self._res, self._actset) = self._ss.derive()
        if self._solver == "direct":
            self._steady_state_vector = (ctmc(create_matrix(self._res)))
        elif self._solver == "sparse":
            self._steady_state_vector = (ctmc_sparse(create_lil_matrix(self._res),
                                         len(self._res)))
        self._vect_names = self._prepare_new_vector(self._res)

    def _prepare_new_vector(self,res):
        vect_names = []
        for key in sorted(res, key=lambda k: res[k][1]):
            vect_names.append(key)
        return vect_names

    def get_vect_names(self):
        return self._vect_names

    def get_steady_state_vector(self):
        return self._steady_state_vector

    def get_actions_throughoutput_from_vector(self, v):
        """ Calculates throughoutput of vector """
        act_vectors = {}
        ret_list = []
        vect_len = len(v)
        for (action,state) in self._actset.keys():
            if action not in act_vectors:
                act_vectors[action] = [0] * vect_len
            act_vectors[action][state-1] = self._actset[ (action, state) ]
        for action in act_vectors.keys():
            ret_list.append( (action, vector_mult(v, act_vectors[action])))
        return ret_list

    def get_actions_throughoutput(self):
        act_vectors = {}
        ret_list = []
        vect_len = len(self._steady_state_vector)
        for (action,state) in self._actset.keys():
            if action not in act_vectors:
                act_vectors[action] = [0] * vect_len
            act_vectors[action][state-1] = self._actset[ (action, state) ]
        for action in act_vectors.keys():
            ret_list.append( (action, vector_mult(self._steady_state_vector, 
                                                  act_vectors[action])))
        return ret_list







#!/usr/bin/env python

from pypepa.parsing.comp_state_space_graph import Transition, ModelSSGraph, ComponentState
from pypepa.derivation.state_space import StateSpace, Operator, Component
from pypepa.parsing.pepa_ast import *
from pypepa.logger import init_log


class PEPATreeWalker():
    """ Various AST Tree methods to generate state spaces
    """
    def __init__(self, rates):
        self._visitstack= []
        self.graph = ModelSSGraph()
        self.log = init_log()
        self.seq_components = {}
        self.shared_actions = {}
        self.operators = []
        self.components = []
        self.ss = StateSpace()
        self.rates = rates

    def prepare(self):
        pass


    def derive_process_state_space(self, node, rates):
        """ Takes node returns state space of a single
            process (here process is P = (a,r).P;
            Assigns rates values
        """
        node = self._visit_tree1(node)
        self._visitstack = []
        self.graph.name = node.process
        self.log.info("Deriving {}".format(node.process))
        self._visit_tree2(node)

    def _name_subtree(self,node):
        """ Names the subtree
        """
        # current = ""
        lcurrent,rcurrent = "",""

        if node.left is not None:
            if node.left.asttype in ("choice", "prefixx"):
                lcurrent += "("
                lcurrent +=  self._name_subtree(node.left)
                lcurrent += ")"
            else:
                lcurrent +=  self._name_subtree(node.left)
        if node.right is not None:
            if node.right.asttype in ("choice", "prefixx"):
                rcurrent += "("
                rcurrent += self._name_subtree(node.right)
                rcurrent += ")"
            else:
                rcurrent += self._name_subtree(node.right)
        return lcurrent + node.data + rcurrent
        # return current


    def derive_systemeq(self, node):
        """ API function for generating state space of a process (node)"""
        self.operators = []
        self.components = []
        self._visit_systemeq(node)
        self.graph.shared_actions = self.shared_actions
        self.ss.operators = self.operators
        self.ss.components = self.components
        return self.ss


    def _visit_systemeq(self, node):
        """ Every procdef in system equation gets offset number
            describing the position in a system equation.
            Additionally every node gets length, which is the sum of
            lengths of its children. Leaves get length 1.
            Operators are added to operators list, whereas components to
            components.
        """
        if node.asttype == "procdef":
            c = Component(self.graph.ss, node.data, len(self.components))
            c.length = 1
            if self.ss.max_length < c.length:
                self.ss.max_length = c.length
            # TODO zmienic, to redundantne i idiotyczne
            c.data = node.data
            node.length = 1
            node.offset = len(self.components)
            self.components.append(c)
        if node.asttype != "procdef":
            c = Operator()
            c.actionset = list(node.actionset) if node.actionset is not None else []
            self.operators.append(c)
        if node.left is not None:
            l = self._visit_systemeq(node.left)
            c.lhs = l
        if node.right is not None:
            r = self._visit_systemeq(node.right)
            c.rhs = r
        if node.asttype != "procdef":
            c.length = l.length + r.length
            if self.ss.max_length < c.length:
                self.ss.max_length = c.length
            node.length = node.left.length + node.right.length
        return c


    def _visit_tree1(self,node):
        """
        Walk the AST tree for the first time to reduce the tree.
        TODO: merge with _visit_tree2
        """
        if node.data == "=":
            node.process = node.left.data
            node.resolved = self._name_subtree(node.right)
            node.left = None
        elif node.data == ".":
            node.action = node.left.action
            # assign numerical value to rate
            if node.left.rate in self.rates:
                node.var_rate = node.left.rate
                node.rate = self.rates[node.left.rate]
            else:
                node.rate = node.left.rate
            node.resolved = self._name_subtree(node.right)
            node.left = None
            if node.right.asttype == "procdef":
                #  the end of the tree
                node.right = None
        elif node.data == "+":
            node.resolved = self._name_subtree(node)
            node.lhs = self._name_subtree(node.left)
            node.rhs = self._name_subtree(node.right)
        elif node.asttype == "procdef":
            pass
        elif node.asttype == "activity":
            pass
        if node.left is not None:
            self._visit_tree1(node.left)
        if node.right is not None:
            self._visit_tree1(node.right)
        return node

    def _visit_tree2(self,node):
        """ Second visiting of the tree, this tree is of elements +.=
            Additionally the function creates graph
        """
        if node.data == "=":
            # create new node, we are in the first node of AST
            compnode = ComponentState()
            compnode.name = node.process
            compnode.resolved = node.resolved
            # adding to ss dict
            self.graph.ss[node.process] = compnode
            # first node for sure
            self.log.debug("(COMPONENT) {} = {}".format(node.process,node.resolved))
            self._visitstack.append(node.process)
        elif node.data == ".":
            trans = Transition(node.action, node.rate, node.resolved)
            if node.var_rate is not None:
                trans.var_rate = node.var_rate
            # add transition to the last state (in the graph)
            self.graph.ss[self._visitstack[-1]].transitions.append( trans )
            self.log.debug("(TR) " + self._visitstack[-1] + " -(" + node.action +","+ node.rate +")-> " + node.resolved)
            # new state again, but if it exists...
            if node.resolved not in self.graph.ss:
                # new state - append
                compnode = ComponentState()
                compnode.resolved = node.resolved
                self.graph.ss[node.resolved] = compnode
#            print("Appending " + node.resolved)
            self._visitstack.append(node.resolved)
        elif node.data == "+":
            pass
        if node.left is not None:
            self._visit_tree2(node.left)
        if node.right is not None:
            self._visit_tree2(node.right)
        if node.data !="+":
            self._visitstack.pop()





#/usr/bin/env python
from pypepa.logger import init_log
from pypepa.parsing.pepa_treewalker import PEPATreeWalker
from pypepa.parsing.comp_state_space_graph import ComponentSSGraph
from pypepa.parsing.parser import PEPAParser
from pypepa.solvers.solution import CTMCSolution
from pypepa.parsing.component_state_visitor import ComponentStateVisitor
import os

class PEPAModel():
    """
        Representation of a final PEPA model, everything needed to derive CTMC
        - state spaces of components that are present in a system equation
        - rate definitions
        - system equation
    """
    def __init__(self, **kwargs):
        """ Create PEPA model instance and fill the fields """
        self.args = kwargs

        # Set up model_filename and model_string depending upon which is
        # is set in the arguments.
        self.model_filename = kwargs.get("file", None)
        self.model_string = kwargs.get("modelstring", None)
        self.name = kwargs.get("name", self.model_filename)
        # In case the name is still None, we set it to a default
        if self.name == None:
          self.name = "model"

        self.processes = {}
        self.systemeq = None
        self.rate_definitions = {}
        self.components = {}
        # from BU alg, get rid of it
        self.tw = None
        self.ss = None
        self.log = init_log()
        self._solver = None
        self.log.info("Starting got args {}".format(kwargs))
        self._parse_model()

    def _parse_model(self):
        """ Parses the model into the abstract syntax tree and sets the name
            of the model. Both of these depend slightly on whether we are
            parsing the model from a string or a file.
        """
        # If self.model_filename has been set then we assume that we should
        # read that in, as the file may have been updated. However we may never
        # have received a model_filename in which case we should use the
        # model_string. If neither of these two things exist we are in trouble.
        if self.model_filename:
            with open(self.model_filename, "r") as f:
                model_string = f.read()
        elif self.model_string:
            model_string = self.model_string
        else:
            raise Exception("No model file or model string present.")
        try:
            parser = PEPAParser()
            (self.processes, self.rate_definitions,
            self.systemeq, self.actions) = parser.parse(model_string)
        except:
            raise


    def get_rates(self):
        return self.rate_definitions

    def derive(self, force=False):
        # Rather than having logic for when you have and have not
        # derived the state space outside of this module, instead we
        # can simply have all procedures which require it call this
        # method. So, this method lazily only calls the statespace
        # if it hasn't been called previously. We have a force argument to
        # force it to be recomputed should we ever require that, but
        # frankly in that case it would easier just to create a new
        # PEPAModel.
        if self.ss is None or force:
            self._prepare_components()

    def recalculate(self, rates=None):
        self._parse_model()
        self._prepare_components(rates)

    def steady_state(self):
        self._derive_steady_state()

    def transient(self, timestop, timestart=0):
        self.derive()
        self.ss.comp_ss = self.tw.graph.ss
        self._solver = CTMCSolution(self.ss, self.args["solver"])
        return self._solver.solve_transient(timestop, timestart)

    def get_steady_state_vector(self):
        return self._solver.get_steady_state_vector()

    def get_state_names(self):
        return self._solver.get_vect_names()

    def get_throughoutput(self):
        return self._solver.get_actions_throughoutput()

    def _derive_steady_state(self):
        """ Derives global state space """
        self.derive()
        self.ss.comp_ss = self.tw.graph.ss
        self._solver = CTMCSolution(self.ss, self.args["solver"])
        self._solver.solve_steady()

    def _prepare_components(self, rateDef=None):
        """ Here ss graphs of every process is derived from AST trees
            as well as state space of components is derived
        """
        if rateDef is None:
            self.tw = PEPATreeWalker(self.rate_definitions)
        else:
            self.log.info("Deriving model with changed rates {}"
                          .format(rateDef))
            self.tw = PEPATreeWalker(rateDef)
        for node in self.processes.values():
            self.tw.derive_process_state_space(node, self.rate_definitions)
        self.ss = self.tw.derive_systemeq(self.systemeq)

    def generate_dots(self, out_dir = "dots"):
        """
        Generates dot files to a specified directory.
        The best application to browse dot files interactively is xdot.
        pip install xdot
        """
        self.log.info("Generating dot files in: %s" % out_dir)
        self.derive()
        visitor = ComponentStateVisitor(self.tw.graph, output_dir = out_dir)
        for comp in set(self.ss.components):
            comptmp = ComponentSSGraph(comp.data)
            self.components[comp.data] = visitor.generate_ss(comp.data,
                                                             comptmp)
        for comp in set(self.ss.components):
            visitor.get_dot(comp.data)








import os
import sys

try:
        from setuptools import setup
except ImportError:
        from distutils.core import setup


setup (
        name = "pypepa",
        version = "0.4.8dev1",
        description = "pypepa is a PEPA library and a toolset for PEPA.",
        long_description = open("README.rst").read(),
        author = "Dariusz Dwornikowski", 
        author_email = "dariusz.dwornikowski@cs.put.poznan.pl",
        url = "https://github.com/tdi/pyPEPA",
        packages = ["pypepa",
                    "pypepa.cli",
                    "pypepa.parsing",
                    "pypepa.derivation",
                    "pypepa.experiments",
                    "pypepa.solvers",
            ],
        package_data={'': ['LICENCE.txt']},
        requires = [
           'numpy (>=1.6.7)',
            'scipy (>=0.10)', 
            'pyparsing (==1.5.7)',
            'matplotlib (>=1.2.0)',
            'colorama (>=0.2.3)',
            ],
        install_requires = ['numpy>=1.6.7', 'scipy>=0.11', 'pyparsing==1.5.6',
                             'matplotlib>=1.2.1'],
        license = "Apache Common 2.0",
        entry_points = {
            'console_scripts': ['pypepa=pypepa.cli.cli:main',]
            },
        classifiers = [
            'Environment :: Console',
            'Development Status :: 4 - Beta',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 2.7',
            'License :: OSI Approved :: Apache Software License',
            'Intended Audience :: Science/Research',
            'Operating System :: POSIX',
            ],
)






#!/usr/bin/env python
import numpy
import scipy
from scipy.sparse import lil_matrix
from scipy.sparse.linalg import spsolve
from scipy.linalg import expm

def ctmc(Q):
    b = numpy.zeros(len(Q), dtype=numpy.float64)
    b[0] = 1
    # normalization
    Q[:,0] = 1
    return numpy.linalg.solve(Q.transpose(),b)

def ctmc_sparse(Q, size):
    b = numpy.zeros(size, dtype=numpy.float64)
    b[0] = 1
    # normalization
    Q[:,0] = 1
    Q =Q.tocsr()
    return spsolve(Q.transpose(),b)

def ctmc_transient(Q, size, tstart, tstop):
    b = numpy.zeros(len(Q), dtype=numpy.float64)
    b[0] = 1
    Qnew = expm(Q*tstop)
    return Qnew[0]

def vector_mult(v1, v2):
    return float(numpy.vdot(numpy.array(v1), numpy.array(v2)))

def create_lil_matrix(res):
    size = len(res)
    QD = numpy.zeros( (size, size), dtype=numpy.float64)
    Q = lil_matrix(QD)
    del QD
    rowsum = 0.0
    for key in sorted(res, key=lambda k: res[k][1]):
        for tos in res[key][0]:
            from_state = int( res[key][1] )
            to_state = int( res[tos[1]][1] )
            rate = float(tos[0])
            Q[ from_state-1 , to_state-1 ] = rate
            rowsum += rate
        Q[ res[key][1]-1 , res[key][1]-1] = -rowsum
        rowsum = 0.0
    # print(Q)
    return Q


def create_matrix(res):
    # create matrix lnum x lnum
    size = len(res)
    Q = numpy.zeros( (size, size), dtype=numpy.float64)
    rowsum = 0.0
    for key in sorted(res, key=lambda k: res[k][1]):
        for tos in res[key][0]:
            from_state = int( res[key][1] )
            to_state = int( res[tos[1]][1] )
            rate = float(tos[0])
            Q[ from_state-1 , to_state-1 ] = rate
            rowsum += rate
        Q[ res[key][1]-1 , res[key][1]-1] = -rowsum
        rowsum = 0.0
    return Q


